/*****************************************************************************
* Product:  Test for "C+" Object Oriented Programming in C
* Version:  3.0
* Updated:  Feb 07, 2005
*
* Copyright (C) 2002-2005 Quantum Leaps. All rights reserved.
*
* This software may be distributed and modified under the terms of the
* BSD open source license, which is provided below.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
* Redistributions of source code must retain the above copyright notice, this
* list of conditions and the following disclaimer.
*
* Redistributions in binary form must reproduce the above copyright notice,
* this list of conditions and the following disclaimer in the documentation
* and/or other materials provided with the distribution.
*
* Neither the name of the Quantum Leaps nor the names of its contributors
* may be used to endorse or promote products derived from this software
* without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL QUANTUM LEAPS OR OTHER CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
* Contact information:
* Quantum Leaps Web site:  http://www.quantum-leaps.com
* Quantum Leaps licensing: http://www.quantum-leaps.com/licensing
* e-mail:                  sales@quatnum-leaps.com
*****************************************************************************/
#include "shape.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/* String class ------------------------------------------------------------*/
void String_ctor1(String *me, char const *str) {
    me->buf__ = (char *)malloc(strlen(str) + 1);
    assert(me->buf__ != (char *)0);
    strcpy(me->buf__, str);
}
/*..........................................................................*/
void String_ctor2(String *me, String *other) {
    String_ctor1(me, String_toChar(other));
}
/*..........................................................................*/
char const *String_toChar(String *me) {
    return me->buf__;
}
/*..........................................................................*/
void String_xtor(String *me) {
    free(me->buf__);                                      /* release buffer */
}

/* Shape class -------------------------------------------------------------*/
BEGIN_VTABLE(Shape, Object)
    VMETHOD(Object, xtor) = (void (*)(Object *))&Shape_xtor_;
    VMETHOD(Shape, area) = (double (*)(Shape *))&Object_abstract;
    VMETHOD(Shape, scale) = (void (*)(Shape *, double))&Object_abstract;
END_VTABLE

/*..........................................................................*/
void Shape_ctor_(Shape *me, char *name) {
    Object_ctor_(&me->super_);                      /* construct superclass */
    VHOOK(Shape);                                       /* hook Shape class */
    String_ctor1(&me->name, name);                      /* construct member */
}
/*..........................................................................*/
void Shape_xtor_(Shape *me) {
    String_xtor(&me->name);                               /* destroy member */
    Object_xtor_(&me->super_);                        /* destroy superclass */
}

/* tests for Shape =========================================================*/
void test_area(Shape *s) {
    assert(Object_IS_KIND_OF(s, Shape));
    printf("name=\"%s\", area()=%.2f, ",
           String_toChar(&s->name),                       /* static binding */
           VCALL(Shape, area, s)END_CALL);               /* dynamic binding */
}
/*..........................................................................*/
void test_scale(Shape *s) {
    double mag = 2.0;
    assert(Object_IS_KIND_OF(s, Shape));
    printf("scale(%.0f), ", mag);
    VCALL(Shape, scale, s), mag END_CALL;                /* dynamic binding */
}


